﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using Infinium.Operation.Contract.Relay;
using QRCoder;

namespace Infinium.Operation.Client.BridgeClient
{
    internal partial class MainUI : System.Windows.Forms.Form
    {
        private WTelegram.Client client;
        private static X509Certificate2 ConnectCertificate;
        private static System.Threading.CancellationTokenSource cancellationTokenSource;
        private static bool IsRunning;
        public MainUI()
        {
            this.InitializeComponent();
            this.Resize += delegate
            {
                if (this.WindowState == System.Windows.Forms.FormWindowState.Minimized && this.tray.Tag == null)
                {
                    tray.ShowBalloonTip(1000);
                    this.Hide();
                }
            };

            this.tray.DoubleClick += delegate
            {
                this.tray.Tag = true;
                this.Show();
                this.WindowState = System.Windows.Forms.FormWindowState.Normal;
                this.tray.Tag = null;
            };

            this.Shown += this.MainUI_Shown;
            this.FormClosing += (s, e) => cancellationTokenSource?.Cancel();
        }

        private void MainUI_Shown(object sender, EventArgs e)
        {
            WTelegram.Helpers.Log = (i, s) => this.InvokeSafe(delegate { this.txtLog.AppendText(s + Environment.NewLine); });

            try
            {
                if (txtApiId.Text.Length == 0 || txtApiHash.Text.Length == 0)
                    return;
                if (!int.TryParse(this.txtApiId.Text, out int apiId))
                    return;
                string sessionPath = Path.GetFullPath("telegram.session");
                if (!File.Exists(sessionPath))
                    return;

                this.Size = new System.Drawing.Size(this.Width, this.Height + this.QRBox.Height);

                client = new WTelegram.Client(apiId, this.txtApiHash.Text, sessionPath);

                //client.LoginUserIfNeeded(null, false).ContinueWith((t) => this.InvokeSafe(() => ContinueLoadAfterTelegramLogin(t)));
                client.ConnectAsync(true).ConfigureAwait(false);

                string coordinatorLogin = Path.GetFullPath("..\\Coordinator\\LastLogin.bin");
                if (!File.Exists(coordinatorLogin))
                    return;

                using (FileStream fs = File.OpenRead(coordinatorLogin))
                {
                    cacheLogin login = ProtoBuf.Serializer.Deserialize<cacheLogin>(fs);
                    if (!login.SignedIn)
                        return;
                    using X509Store store = new X509Store(StoreName.My, StoreLocation.CurrentUser);
                    store.Open(OpenFlags.ReadOnly | OpenFlags.OpenExistingOnly);
                    ConnectCertificate = store.Certificates.FirstOrDefault(x => x.GetCertHash().AsSpan().SequenceEqual(login.LastThumbprint));
                    if (ConnectCertificate == null)
                        throw new OperationCanceledException("Unable to find certificate for last Coordinator login");
                    if (ConnectCertificate.NotAfter < DateTime.Now)
                        throw new OperationCanceledException("Connect certificate expired");
                }

                IsRunning = true;

                this.btnConnect.Enabled = false;
                Task.Factory.StartNew(() =>
                {
                    while (IsRunning)
                    {
                        this.Listen().Wait();
                        Task.Delay(5000).Wait();
                    }
                });
            }
            catch (OperationCanceledException ocex) { System.Windows.Forms.MessageBox.Show(this, ocex.Message); }
            catch (Exception ex) { System.Windows.Forms.MessageBox.Show(this, ex.ToString()); }
        }

        private void InvokeSafe(Action action)
            => this.Invoke((System.Windows.Forms.MethodInvoker)delegate { action(); });

        private void ContinueLoadAfterTelegramLogin(Task<TL.User> taskLogin)
        {
            this.btnConnect.Enabled = taskLogin.IsFaulted;
            if (!taskLogin.IsFaulted)
                this.btnLogin.Text = "Telegram logged in";
            else
            {
                this.Size = new System.Drawing.Size(this.Width, this.Height - this.QRBox.Height);
                this.txtLog.SendToBack();
                tray.ShowBalloonTip(1000
                    , "Infinium Bridge Client"
                    , "Telegram loggin failed when resuming client session"
                    , System.Windows.Forms.ToolTipIcon.Warning);
            }
        }

        private void btnLogin_Click(object sender, EventArgs e)
        {
            try
            {
                if (this.txtApiId.Text.Length == 0)
                    throw new ArgumentException("Please fill api_id");
                if (this.txtApiHash.Text.Length == 0)
                    throw new ArgumentException("Please fill api_hash");
                int apiId = int.Parse(this.txtApiId.Text);
                string sessionPath = Path.GetFullPath("telegram.session");
                client = new WTelegram.Client(apiId, this.txtApiHash.Text, sessionPath);

                this.Size = new System.Drawing.Size(this.Width, this.Height + this.QRBox.Height);
                this.Update();
                Task<TL.User> taskLogin = client.LoginWithQRCode(this.showQRCode);
                taskLogin.ContinueWith(t => this.InvokeSafe(() => this.HandleTelegramLoginResult(t)));
                this.btnLogin.Enabled = false;
                this.txtLog.SendToBack();
            }
            catch (ArgumentException anex) { System.Windows.Forms.MessageBox.Show(this, anex.Message); }
            catch (Exception ex) { System.Windows.Forms.MessageBox.Show(this, ex.ToString()); }
        }

        private void HandleTelegramLoginResult(Task<TL.User> taskLogin)
        {
            var user = taskLogin.Result;
            if (user.id != 0)
            {
                this.QRBox.Visible = false;
                this.btnLogin.Text = "Telegram logged in";
                this.tray.ShowBalloonTip(500
                    , "Infinium Bridge Client"
                    , $"Telegram logged in as [{user.MainUsername}]"
                    , System.Windows.Forms.ToolTipIcon.Info);
            }
            else
                this.btnLogin.Enabled = true;

            this.Size = new System.Drawing.Size(this.Width, this.Height - this.QRBox.Height);
        }

        private void showQRCode(string str)
        {
            using (QRCodeGenerator qrGenerator = new QRCodeGenerator())
            using (QRCodeData qrCodeData = qrGenerator.CreateQrCode(str, QRCodeGenerator.ECCLevel.Q))
            using (PngByteQRCode qrCode = new PngByteQRCode(qrCodeData))
            {
                byte[] qrCodeImage = qrCode.GetGraphic(20);
                using (System.IO.MemoryStream ms = new System.IO.MemoryStream(qrCodeImage))
                {
                    System.Drawing.Image image = System.Drawing.Image.FromStream(ms);
                    QRBox.Image = image;
                }
            }
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            this.btnConnect.Enabled = false;
            IsRunning = true;
            Task.Factory.StartNew(() =>
            {
                while (IsRunning)
                    this.Listen().Wait();
            }, TaskCreationOptions.PreferFairness | TaskCreationOptions.LongRunning);
        }

        private async Task Listen()
        {
            try
            {
                cancellationTokenSource = new System.Threading.CancellationTokenSource();
                ProtoBuf.Grpc.CallContext callContext = new ProtoBuf.Grpc.CallContext(new Grpc.Core.CallOptions(cancellationToken: cancellationTokenSource.Token));
                GameChat.ClientStatus status = new GameChat.ClientStatus(client.UserId, 0, true);
                if (!IsRunning)
                    return;

                await foreach (var chat in ClientRegis.GetService<IRelayGameChat>().SubscribeRelay(status, callContext))
                    try
                    {
                        if (chat.ChannelType == 0xFF)
                        {
                            //Signal connected;
                            this.Invoke((System.Windows.Forms.MethodInvoker)delegate { this.btnConnect.Text = "Connected"; });
                            return;
                        }

                        ChatGroupType type = (ChatGroupType)chat.ChannelType;
                        TL.InputPeer peer;
                        if (type == ChatGroupType.GuildChat)
                            peer = new TL.InputPeerChat(chat.ChannelId);
                        else if (type == ChatGroupType.GuildChannel)
                            peer = new TL.InputPeerChannel(chat.ChannelId, chat.ChannelHash);
                        else
                            continue;

                        TL.MessageEntity[] entities = chat.Entities?.Select(x => new TL.MessageEntityTextUrl()
                        {
                            length = x.Length,
                            offset = x.Offset,
                            url = x.Url,
                        }).ToArray();

                        _ = client.SendMessageAsync(peer, chat.Text, entities: entities);
                    }
                    catch { }
            }
            catch (Grpc.Core.RpcException rex)
            {
                if (rex.StatusCode != Grpc.Core.StatusCode.Cancelled)
                    Task.WaitAll(Task.Delay(30000));
            }
            catch (Exception)
            {
                Task.WaitAll(Task.Delay(10000));
            }
        }

        private void chkShowLog_CheckedChanged(object sender, EventArgs e)
        {
            if (chkShowLog.Checked)
            {
                if (this.Size.Height < this.txtLog.Size.Height)
                    this.Size = new System.Drawing.Size(this.Width, this.Height + this.QRBox.Height);
                this.txtLog.BringToFront();
            }
            else
                this.txtLog.SendToBack();
        }
    }
}
